-file("/usr/local/lib/erlang/lib/parsetools-2.1.6/include/leexinc.hrl", 0).
%% The source of this file is part of leex distribution, as such it
%% has the same Copyright as the other files in the leex
%% distribution. The Copyright is defined in the accompanying file
%% COPYRIGHT. However, the resultant scanner generated by leex is the
%% property of the creator of the scanner and is not covered by that
%% Copyright.

-module(cuneiform_scan).

-export([string/1,string/2,token/2,token/3,tokens/2,tokens/3]).
-export([format_error/1]).

%% User code. This is placed here to allow extra attributes.
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 175).

-export( [yyrev/2] ).

-spec trim_body( S :: string() ) -> string().

trim_body( S )
when is_list( S ), length( S ) >= 4 ->
  string:substr( S, 3, length( S )-4 ).


-spec trim_lit( S :: string() ) -> string().

trim_lit( S )
when is_list( S ), length( S ) >= 2 ->
  string:substr( S, 2, length( S )-2 ).
-file("/usr/local/lib/erlang/lib/parsetools-2.1.6/include/leexinc.hrl", 14).

format_error({illegal,S}) -> ["illegal characters ",io_lib:write_string(S)];
format_error({user,S}) -> S.

string(String) -> string(String, 1).

string(String, Line) -> string(String, Line, String, []).

%% string(InChars, Line, TokenChars, Tokens) ->
%% {ok,Tokens,Line} | {error,ErrorInfo,Line}.
%% Note the line number going into yystate, L0, is line of token
%% start while line number returned is line of token end. We want line
%% of token start.

string([], L, [], Ts) ->                     % No partial tokens!
    {ok,yyrev(Ts),L};
string(Ics0, L0, Tcs, Ts) ->
    case yystate(yystate(), Ics0, L0, 0, reject, 0) of
        {A,Alen,Ics1,L1} ->                  % Accepting end state
            string_cont(Ics1, L1, yyaction(A, Alen, Tcs, L0), Ts);
        {A,Alen,Ics1,L1,_S1} ->              % Accepting transistion state
            string_cont(Ics1, L1, yyaction(A, Alen, Tcs, L0), Ts);
        {reject,_Alen,Tlen,_Ics1,L1,_S1} ->  % After a non-accepting state
            {error,{L0,?MODULE,{illegal,yypre(Tcs, Tlen+1)}},L1};
        {A,Alen,Tlen,_Ics1,L1,_S1} ->
            Tcs1 = yysuf(Tcs, Alen),
            L2 = adjust_line(Tlen, Alen, Tcs1, L1),
            string_cont(Tcs1, L2, yyaction(A, Alen, Tcs, L0), Ts)
    end.

%% string_cont(RestChars, Line, Token, Tokens)
%% Test for and remove the end token wrapper. Push back characters
%% are prepended to RestChars.

-dialyzer({nowarn_function, string_cont/4}).

string_cont(Rest, Line, {token,T}, Ts) ->
    string(Rest, Line, Rest, [T|Ts]);
string_cont(Rest, Line, {token,T,Push}, Ts) ->
    NewRest = Push ++ Rest,
    string(NewRest, Line, NewRest, [T|Ts]);
string_cont(Rest, Line, {end_token,T}, Ts) ->
    string(Rest, Line, Rest, [T|Ts]);
string_cont(Rest, Line, {end_token,T,Push}, Ts) ->
    NewRest = Push ++ Rest,
    string(NewRest, Line, NewRest, [T|Ts]);
string_cont(Rest, Line, skip_token, Ts) ->
    string(Rest, Line, Rest, Ts);
string_cont(Rest, Line, {skip_token,Push}, Ts) ->
    NewRest = Push ++ Rest,
    string(NewRest, Line, NewRest, Ts);
string_cont(_Rest, Line, {error,S}, _Ts) ->
    {error,{Line,?MODULE,{user,S}},Line}.

%% token(Continuation, Chars) ->
%% token(Continuation, Chars, Line) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.
%% Must be careful when re-entering to append the latest characters to the
%% after characters in an accept. The continuation is:
%% {token,State,CurrLine,TokenChars,TokenLen,TokenLine,AccAction,AccLen}

token(Cont, Chars) -> token(Cont, Chars, 1).

token([], Chars, Line) ->
    token(yystate(), Chars, Line, Chars, 0, Line, reject, 0);
token({token,State,Line,Tcs,Tlen,Tline,Action,Alen}, Chars, _) ->
    token(State, Chars, Line, Tcs ++ Chars, Tlen, Tline, Action, Alen).

%% token(State, InChars, Line, TokenChars, TokenLen, TokenLine,
%% AcceptAction, AcceptLen) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.
%% The argument order is chosen to be more efficient.

token(S0, Ics0, L0, Tcs, Tlen0, Tline, A0, Alen0) ->
    case yystate(S0, Ics0, L0, Tlen0, A0, Alen0) of
        %% Accepting end state, we have a token.
        {A1,Alen1,Ics1,L1} ->
            token_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline));
        %% Accepting transition state, can take more chars.
        {A1,Alen1,[],L1,S1} ->                  % Need more chars to check
            {more,{token,S1,L1,Tcs,Alen1,Tline,A1,Alen1}};
        {A1,Alen1,Ics1,L1,_S1} ->               % Take what we got
            token_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline));
        %% After a non-accepting state, maybe reach accept state later.
        {A1,Alen1,Tlen1,[],L1,S1} ->            % Need more chars to check
            {more,{token,S1,L1,Tcs,Tlen1,Tline,A1,Alen1}};
        {reject,_Alen1,Tlen1,eof,L1,_S1} ->     % No token match
            %% Check for partial token which is error.
            Ret = if Tlen1 > 0 -> {error,{Tline,?MODULE,
                                          %% Skip eof tail in Tcs.
                                          {illegal,yypre(Tcs, Tlen1)}},L1};
                     true -> {eof,L1}
                  end,
            {done,Ret,eof};
        {reject,_Alen1,Tlen1,Ics1,L1,_S1} ->    % No token match
            Error = {Tline,?MODULE,{illegal,yypre(Tcs, Tlen1+1)}},
            {done,{error,Error,L1},Ics1};
        {A1,Alen1,Tlen1,_Ics1,L1,_S1} ->       % Use last accept match
            Tcs1 = yysuf(Tcs, Alen1),
            L2 = adjust_line(Tlen1, Alen1, Tcs1, L1),
            token_cont(Tcs1, L2, yyaction(A1, Alen1, Tcs, Tline))
    end.

%% token_cont(RestChars, Line, Token)
%% If we have a token or error then return done, else if we have a
%% skip_token then continue.

-dialyzer({nowarn_function, token_cont/3}).

token_cont(Rest, Line, {token,T}) ->
    {done,{ok,T,Line},Rest};
token_cont(Rest, Line, {token,T,Push}) ->
    NewRest = Push ++ Rest,
    {done,{ok,T,Line},NewRest};
token_cont(Rest, Line, {end_token,T}) ->
    {done,{ok,T,Line},Rest};
token_cont(Rest, Line, {end_token,T,Push}) ->
    NewRest = Push ++ Rest,
    {done,{ok,T,Line},NewRest};
token_cont(Rest, Line, skip_token) ->
    token(yystate(), Rest, Line, Rest, 0, Line, reject, 0);
token_cont(Rest, Line, {skip_token,Push}) ->
    NewRest = Push ++ Rest,
    token(yystate(), NewRest, Line, NewRest, 0, Line, reject, 0);
token_cont(Rest, Line, {error,S}) ->
    {done,{error,{Line,?MODULE,{user,S}},Line},Rest}.

%% tokens(Continuation, Chars, Line) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.
%% Must be careful when re-entering to append the latest characters to the
%% after characters in an accept. The continuation is:
%% {tokens,State,CurrLine,TokenChars,TokenLen,TokenLine,Tokens,AccAction,AccLen}
%% {skip_tokens,State,CurrLine,TokenChars,TokenLen,TokenLine,Error,AccAction,AccLen}

tokens(Cont, Chars) -> tokens(Cont, Chars, 1).

tokens([], Chars, Line) ->
    tokens(yystate(), Chars, Line, Chars, 0, Line, [], reject, 0);
tokens({tokens,State,Line,Tcs,Tlen,Tline,Ts,Action,Alen}, Chars, _) ->
    tokens(State, Chars, Line, Tcs ++ Chars, Tlen, Tline, Ts, Action, Alen);
tokens({skip_tokens,State,Line,Tcs,Tlen,Tline,Error,Action,Alen}, Chars, _) ->
    skip_tokens(State, Chars, Line, Tcs ++ Chars, Tlen, Tline, Error, Action, Alen).

%% tokens(State, InChars, Line, TokenChars, TokenLen, TokenLine, Tokens,
%% AcceptAction, AcceptLen) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.

tokens(S0, Ics0, L0, Tcs, Tlen0, Tline, Ts, A0, Alen0) ->
    case yystate(S0, Ics0, L0, Tlen0, A0, Alen0) of
        %% Accepting end state, we have a token.
        {A1,Alen1,Ics1,L1} ->
            tokens_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline), Ts);
        %% Accepting transition state, can take more chars.
        {A1,Alen1,[],L1,S1} ->                  % Need more chars to check
            {more,{tokens,S1,L1,Tcs,Alen1,Tline,Ts,A1,Alen1}};
        {A1,Alen1,Ics1,L1,_S1} ->               % Take what we got
            tokens_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline), Ts);
        %% After a non-accepting state, maybe reach accept state later.
        {A1,Alen1,Tlen1,[],L1,S1} ->            % Need more chars to check
            {more,{tokens,S1,L1,Tcs,Tlen1,Tline,Ts,A1,Alen1}};
        {reject,_Alen1,Tlen1,eof,L1,_S1} ->     % No token match
            %% Check for partial token which is error, no need to skip here.
            Ret = if Tlen1 > 0 -> {error,{Tline,?MODULE,
                                          %% Skip eof tail in Tcs.
                                          {illegal,yypre(Tcs, Tlen1)}},L1};
                     Ts == [] -> {eof,L1};
                     true -> {ok,yyrev(Ts),L1}
                  end,
            {done,Ret,eof};
        {reject,_Alen1,Tlen1,_Ics1,L1,_S1} ->
            %% Skip rest of tokens.
            Error = {L1,?MODULE,{illegal,yypre(Tcs, Tlen1+1)}},
            skip_tokens(yysuf(Tcs, Tlen1+1), L1, Error);
        {A1,Alen1,Tlen1,_Ics1,L1,_S1} ->
            Token = yyaction(A1, Alen1, Tcs, Tline),
            Tcs1 = yysuf(Tcs, Alen1),
            L2 = adjust_line(Tlen1, Alen1, Tcs1, L1),
            tokens_cont(Tcs1, L2, Token, Ts)
    end.

%% tokens_cont(RestChars, Line, Token, Tokens)
%% If we have an end_token or error then return done, else if we have
%% a token then save it and continue, else if we have a skip_token
%% just continue.

-dialyzer({nowarn_function, tokens_cont/4}).

tokens_cont(Rest, Line, {token,T}, Ts) ->
    tokens(yystate(), Rest, Line, Rest, 0, Line, [T|Ts], reject, 0);
tokens_cont(Rest, Line, {token,T,Push}, Ts) ->
    NewRest = Push ++ Rest,
    tokens(yystate(), NewRest, Line, NewRest, 0, Line, [T|Ts], reject, 0);
tokens_cont(Rest, Line, {end_token,T}, Ts) ->
    {done,{ok,yyrev(Ts, [T]),Line},Rest};
tokens_cont(Rest, Line, {end_token,T,Push}, Ts) ->
    NewRest = Push ++ Rest,
    {done,{ok,yyrev(Ts, [T]),Line},NewRest};
tokens_cont(Rest, Line, skip_token, Ts) ->
    tokens(yystate(), Rest, Line, Rest, 0, Line, Ts, reject, 0);
tokens_cont(Rest, Line, {skip_token,Push}, Ts) ->
    NewRest = Push ++ Rest,
    tokens(yystate(), NewRest, Line, NewRest, 0, Line, Ts, reject, 0);
tokens_cont(Rest, Line, {error,S}, _Ts) ->
    skip_tokens(Rest, Line, {Line,?MODULE,{user,S}}).

%%skip_tokens(InChars, Line, Error) -> {done,{error,Error,Line},Ics}.
%% Skip tokens until an end token, junk everything and return the error.

skip_tokens(Ics, Line, Error) ->
    skip_tokens(yystate(), Ics, Line, Ics, 0, Line, Error, reject, 0).

%% skip_tokens(State, InChars, Line, TokenChars, TokenLen, TokenLine, Tokens,
%% AcceptAction, AcceptLen) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.

skip_tokens(S0, Ics0, L0, Tcs, Tlen0, Tline, Error, A0, Alen0) ->
    case yystate(S0, Ics0, L0, Tlen0, A0, Alen0) of
        {A1,Alen1,Ics1,L1} ->                  % Accepting end state
            skip_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline), Error);
        {A1,Alen1,[],L1,S1} ->                 % After an accepting state
            {more,{skip_tokens,S1,L1,Tcs,Alen1,Tline,Error,A1,Alen1}};
        {A1,Alen1,Ics1,L1,_S1} ->
            skip_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline), Error);
        {A1,Alen1,Tlen1,[],L1,S1} ->           % After a non-accepting state
            {more,{skip_tokens,S1,L1,Tcs,Tlen1,Tline,Error,A1,Alen1}};
        {reject,_Alen1,_Tlen1,eof,L1,_S1} ->
            {done,{error,Error,L1},eof};
        {reject,_Alen1,Tlen1,_Ics1,L1,_S1} ->
            skip_tokens(yysuf(Tcs, Tlen1+1), L1, Error);
        {A1,Alen1,Tlen1,_Ics1,L1,_S1} ->
            Token = yyaction(A1, Alen1, Tcs, Tline),
            Tcs1 = yysuf(Tcs, Alen1),
            L2 = adjust_line(Tlen1, Alen1, Tcs1, L1),
            skip_cont(Tcs1, L2, Token, Error)
    end.

%% skip_cont(RestChars, Line, Token, Error)
%% Skip tokens until we have an end_token or error then return done
%% with the original rror.

-dialyzer({nowarn_function, skip_cont/4}).

skip_cont(Rest, Line, {token,_T}, Error) ->
    skip_tokens(yystate(), Rest, Line, Rest, 0, Line, Error, reject, 0);
skip_cont(Rest, Line, {token,_T,Push}, Error) ->
    NewRest = Push ++ Rest,
    skip_tokens(yystate(), NewRest, Line, NewRest, 0, Line, Error, reject, 0);
skip_cont(Rest, Line, {end_token,_T}, Error) ->
    {done,{error,Error,Line},Rest};
skip_cont(Rest, Line, {end_token,_T,Push}, Error) ->
    NewRest = Push ++ Rest,
    {done,{error,Error,Line},NewRest};
skip_cont(Rest, Line, skip_token, Error) ->
    skip_tokens(yystate(), Rest, Line, Rest, 0, Line, Error, reject, 0);
skip_cont(Rest, Line, {skip_token,Push}, Error) ->
    NewRest = Push ++ Rest,
    skip_tokens(yystate(), NewRest, Line, NewRest, 0, Line, Error, reject, 0);
skip_cont(Rest, Line, {error,_S}, Error) ->
    skip_tokens(yystate(), Rest, Line, Rest, 0, Line, Error, reject, 0).

-compile({nowarn_unused_function, [yyrev/1, yyrev/2, yypre/2, yysuf/2]}).

yyrev(List) -> lists:reverse(List).
yyrev(List, Tail) -> lists:reverse(List, Tail).
yypre(List, N) -> lists:sublist(List, N).
yysuf(List, N) -> lists:nthtail(N, List).

%% adjust_line(TokenLength, AcceptLength, Chars, Line) -> NewLine
%% Make sure that newlines in Chars are not counted twice.
%% Line has been updated with respect to newlines in the prefix of
%% Chars consisting of (TokenLength - AcceptLength) characters.

-compile({nowarn_unused_function, adjust_line/4}).

adjust_line(N, N, _Cs, L) -> L;
adjust_line(T, A, [$\n|Cs], L) ->
    adjust_line(T-1, A, Cs, L-1);
adjust_line(T, A, [_|Cs], L) ->
    adjust_line(T-1, A, Cs, L).

%% yystate() -> InitialState.
%% yystate(State, InChars, Line, CurrTokLen, AcceptAction, AcceptLen) ->
%% {Action, AcceptLen, RestChars, Line} |
%% {Action, AcceptLen, RestChars, Line, State} |
%% {reject, AcceptLen, CurrTokLen, RestChars, Line, State} |
%% {Action, AcceptLen, CurrTokLen, RestChars, Line, State}.
%% Generated state transition functions. The non-accepting end state
%% return signal either an unrecognised character or end of current
%% input.

-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.erl", 322).
yystate() -> 138.

yystate(141, [116|Ics], Line, Tlen, _, _) ->
    yystate(139, Ics, Line, Tlen+1, 50, Tlen);
yystate(141, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(141, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(141, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(141, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(141, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(141, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(141, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(141, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,141};
yystate(140, Ics, Line, Tlen, _, _) ->
    {56,Tlen,Ics,Line};
yystate(139, [104|Ics], Line, Tlen, _, _) ->
    yystate(135, Ics, Line, Tlen+1, 50, Tlen);
yystate(139, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(139, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(139, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(139, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(139, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(139, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 103 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(139, [C|Ics], Line, Tlen, _, _) when C >= 105, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(139, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,139};
yystate(138, [125|Ics], Line, Tlen, Action, Alen) ->
    yystate(134, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [124|Ics], Line, Tlen, Action, Alen) ->
    yystate(130, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [123|Ics], Line, Tlen, Action, Alen) ->
    yystate(126, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [116|Ics], Line, Tlen, Action, Alen) ->
    yystate(122, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [111|Ics], Line, Tlen, Action, Alen) ->
    yystate(90, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [110|Ics], Line, Tlen, Action, Alen) ->
    yystate(82, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [109|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [108|Ics], Line, Tlen, Action, Alen) ->
    yystate(70, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [106|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [107|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [105|Ics], Line, Tlen, Action, Alen) ->
    yystate(58, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [103|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [104|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [102|Ics], Line, Tlen, Action, Alen) ->
    yystate(10, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [101|Ics], Line, Tlen, Action, Alen) ->
    yystate(25, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [100|Ics], Line, Tlen, Action, Alen) ->
    yystate(65, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [98|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [99|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [97|Ics], Line, Tlen, Action, Alen) ->
    yystate(81, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [93|Ics], Line, Tlen, Action, Alen) ->
    yystate(93, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [91|Ics], Line, Tlen, Action, Alen) ->
    yystate(97, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [83|Ics], Line, Tlen, Action, Alen) ->
    yystate(101, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [82|Ics], Line, Tlen, Action, Alen) ->
    yystate(113, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [81|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [80|Ics], Line, Tlen, Action, Alen) ->
    yystate(137, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [79|Ics], Line, Tlen, Action, Alen) ->
    yystate(111, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [78|Ics], Line, Tlen, Action, Alen) ->
    yystate(87, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [77|Ics], Line, Tlen, Action, Alen) ->
    yystate(75, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [75|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [76|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [74|Ics], Line, Tlen, Action, Alen) ->
    yystate(51, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [70|Ics], Line, Tlen, Action, Alen) ->
    yystate(35, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [69|Ics], Line, Tlen, Action, Alen) ->
    yystate(11, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [67|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [68|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [66|Ics], Line, Tlen, Action, Alen) ->
    yystate(12, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [65|Ics], Line, Tlen, Action, Alen) ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [62|Ics], Line, Tlen, Action, Alen) ->
    yystate(40, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [61|Ics], Line, Tlen, Action, Alen) ->
    yystate(48, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [60|Ics], Line, Tlen, Action, Alen) ->
    yystate(56, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [59|Ics], Line, Tlen, Action, Alen) ->
    yystate(64, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [58|Ics], Line, Tlen, Action, Alen) ->
    yystate(68, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [48|Ics], Line, Tlen, Action, Alen) ->
    yystate(76, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [45|Ics], Line, Tlen, Action, Alen) ->
    yystate(80, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [44|Ics], Line, Tlen, Action, Alen) ->
    yystate(88, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [43|Ics], Line, Tlen, Action, Alen) ->
    yystate(92, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [42|Ics], Line, Tlen, Action, Alen) ->
    yystate(96, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [41|Ics], Line, Tlen, Action, Alen) ->
    yystate(112, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [40|Ics], Line, Tlen, Action, Alen) ->
    yystate(116, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [39|Ics], Line, Tlen, Action, Alen) ->
    yystate(124, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [37|Ics], Line, Tlen, Action, Alen) ->
    yystate(128, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [35|Ics], Line, Tlen, Action, Alen) ->
    yystate(128, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [34|Ics], Line, Tlen, Action, Alen) ->
    yystate(136, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [10|Ics], Line, Tlen, Action, Alen) ->
    yystate(140, Ics, Line+1, Tlen+1, Action, Alen);
yystate(138, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(140, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 32 ->
    yystate(140, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [C|Ics], Line, Tlen, Action, Alen) when C >= 49, C =< 57 ->
    yystate(72, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [C|Ics], Line, Tlen, Action, Alen) when C >= 71, C =< 73 ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [C|Ics], Line, Tlen, Action, Alen) when C >= 84, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [C|Ics], Line, Tlen, Action, Alen) when C >= 112, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, [C|Ics], Line, Tlen, Action, Alen) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(138, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,138};
yystate(137, [122|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [121|Ics], Line, Tlen, _, _) ->
    yystate(141, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [101|Ics], Line, Tlen, _, _) ->
    yystate(123, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 120 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(137, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,137};
yystate(136, [34|Ics], Line, Tlen, Action, Alen) ->
    yystate(132, Ics, Line, Tlen+1, Action, Alen);
yystate(136, [10|Ics], Line, Tlen, Action, Alen) ->
    yystate(136, Ics, Line+1, Tlen+1, Action, Alen);
yystate(136, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(136, Ics, Line, Tlen+1, Action, Alen);
yystate(136, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 33 ->
    yystate(136, Ics, Line, Tlen+1, Action, Alen);
yystate(136, [C|Ics], Line, Tlen, Action, Alen) when C >= 35 ->
    yystate(136, Ics, Line, Tlen+1, Action, Alen);
yystate(136, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,136};
yystate(135, [111|Ics], Line, Tlen, _, _) ->
    yystate(131, Ics, Line, Tlen+1, 50, Tlen);
yystate(135, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(135, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(135, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(135, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(135, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(135, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(135, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(135, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,135};
yystate(134, Ics, Line, Tlen, _, _) ->
    {41,Tlen,Ics,Line};
yystate(133, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 8, Tlen);
yystate(133, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 8, Tlen);
yystate(133, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 8, Tlen);
yystate(133, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 8, Tlen);
yystate(133, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 8, Tlen);
yystate(133, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 8, Tlen);
yystate(133, Ics, Line, Tlen, _, _) ->
    {8,Tlen,Ics,Line,133};
yystate(132, Ics, Line, Tlen, _, _) ->
    {52,Tlen,Ics,Line};
yystate(131, [110|Ics], Line, Tlen, _, _) ->
    yystate(127, Ics, Line, Tlen+1, 50, Tlen);
yystate(131, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(131, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(131, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(131, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(131, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(131, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(131, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(131, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,131};
yystate(130, Ics, Line, Tlen, _, _) ->
    {15,Tlen,Ics,Line};
yystate(129, [116|Ics], Line, Tlen, _, _) ->
    yystate(133, Ics, Line, Tlen+1, 50, Tlen);
yystate(129, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(129, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(129, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(129, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(129, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(129, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(129, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(129, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,129};
yystate(128, [C|Ics], Line, Tlen, _, _) when C >= 0, C =< 9 ->
    yystate(128, Ics, Line, Tlen+1, 55, Tlen);
yystate(128, [C|Ics], Line, Tlen, _, _) when C >= 11 ->
    yystate(128, Ics, Line, Tlen+1, 55, Tlen);
yystate(128, Ics, Line, Tlen, _, _) ->
    {55,Tlen,Ics,Line,128};
yystate(127, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 6, Tlen);
yystate(127, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 6, Tlen);
yystate(127, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 6, Tlen);
yystate(127, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 6, Tlen);
yystate(127, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 6, Tlen);
yystate(127, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 6, Tlen);
yystate(127, Ics, Line, Tlen, _, _) ->
    {6,Tlen,Ics,Line,127};
yystate(126, Ics, Line, Tlen, _, _) ->
    {34,Tlen,Ics,Line};
yystate(125, [101|Ics], Line, Tlen, _, _) ->
    yystate(129, Ics, Line, Tlen+1, 50, Tlen);
yystate(125, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(125, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(125, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(125, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(125, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(125, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(125, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(125, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,125};
yystate(124, [39|Ics], Line, Tlen, Action, Alen) ->
    yystate(120, Ics, Line, Tlen+1, Action, Alen);
yystate(124, [10|Ics], Line, Tlen, Action, Alen) ->
    yystate(124, Ics, Line+1, Tlen+1, Action, Alen);
yystate(124, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(124, Ics, Line, Tlen+1, Action, Alen);
yystate(124, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 38 ->
    yystate(124, Ics, Line, Tlen+1, Action, Alen);
yystate(124, [C|Ics], Line, Tlen, Action, Alen) when C >= 40 ->
    yystate(124, Ics, Line, Tlen+1, Action, Alen);
yystate(124, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,124};
yystate(123, [114|Ics], Line, Tlen, _, _) ->
    yystate(119, Ics, Line, Tlen+1, 50, Tlen);
yystate(123, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(123, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(123, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(123, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(123, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(123, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(123, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(123, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,123};
yystate(122, [114|Ics], Line, Tlen, _, _) ->
    yystate(118, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [104|Ics], Line, Tlen, _, _) ->
    yystate(106, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 103 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [C|Ics], Line, Tlen, _, _) when C >= 105, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(122, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,122};
yystate(121, [107|Ics], Line, Tlen, _, _) ->
    yystate(125, Ics, Line, Tlen+1, 50, Tlen);
yystate(121, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(121, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(121, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(121, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(121, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(121, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 106 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(121, [C|Ics], Line, Tlen, _, _) when C >= 108, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(121, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,121};
yystate(120, Ics, Line, Tlen, _, _) ->
    {53,Tlen,Ics,Line};
yystate(119, [108|Ics], Line, Tlen, _, _) ->
    yystate(115, Ics, Line, Tlen+1, 50, Tlen);
yystate(119, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(119, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(119, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(119, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(119, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(119, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(119, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(119, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,119};
yystate(118, [117|Ics], Line, Tlen, _, _) ->
    yystate(114, Ics, Line, Tlen+1, 50, Tlen);
yystate(118, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(118, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(118, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(118, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(118, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(118, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 116 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(118, [C|Ics], Line, Tlen, _, _) when C >= 118, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(118, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,118};
yystate(117, [99|Ics], Line, Tlen, _, _) ->
    yystate(121, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, [97|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, [98|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, [C|Ics], Line, Tlen, _, _) when C >= 100, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(117, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,117};
yystate(116, Ics, Line, Tlen, _, _) ->
    {35,Tlen,Ics,Line};
yystate(115, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 5, Tlen);
yystate(115, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 5, Tlen);
yystate(115, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 5, Tlen);
yystate(115, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 5, Tlen);
yystate(115, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 5, Tlen);
yystate(115, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 5, Tlen);
yystate(115, Ics, Line, Tlen, _, _) ->
    {5,Tlen,Ics,Line,115};
yystate(114, [101|Ics], Line, Tlen, _, _) ->
    yystate(110, Ics, Line, Tlen+1, 50, Tlen);
yystate(114, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(114, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(114, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(114, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(114, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(114, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(114, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(114, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,114};
yystate(113, [97|Ics], Line, Tlen, _, _) ->
    yystate(117, Ics, Line, Tlen+1, 7, Tlen);
yystate(113, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 7, Tlen);
yystate(113, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 7, Tlen);
yystate(113, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 7, Tlen);
yystate(113, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 7, Tlen);
yystate(113, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 7, Tlen);
yystate(113, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 7, Tlen);
yystate(113, Ics, Line, Tlen, _, _) ->
    {7,Tlen,Ics,Line,113};
yystate(112, Ics, Line, Tlen, _, _) ->
    {42,Tlen,Ics,Line};
yystate(111, [99|Ics], Line, Tlen, _, _) ->
    yystate(107, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, [97|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, [98|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, [C|Ics], Line, Tlen, _, _) when C >= 100, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(111, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,111};
yystate(110, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 47, Tlen);
yystate(110, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 47, Tlen);
yystate(110, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 47, Tlen);
yystate(110, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 47, Tlen);
yystate(110, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 47, Tlen);
yystate(110, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 47, Tlen);
yystate(110, Ics, Line, Tlen, _, _) ->
    {47,Tlen,Ics,Line,110};
yystate(109, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 9, Tlen);
yystate(109, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 9, Tlen);
yystate(109, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 9, Tlen);
yystate(109, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 9, Tlen);
yystate(109, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 9, Tlen);
yystate(109, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 9, Tlen);
yystate(109, Ics, Line, Tlen, _, _) ->
    {9,Tlen,Ics,Line,109};
yystate(108, [125|Ics], Line, Tlen, Action, Alen) ->
    yystate(100, Ics, Line, Tlen+1, Action, Alen);
yystate(108, [10|Ics], Line, Tlen, Action, Alen) ->
    yystate(108, Ics, Line+1, Tlen+1, Action, Alen);
yystate(108, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(108, Ics, Line, Tlen+1, Action, Alen);
yystate(108, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 124 ->
    yystate(108, Ics, Line, Tlen+1, Action, Alen);
yystate(108, [C|Ics], Line, Tlen, Action, Alen) when C >= 126 ->
    yystate(108, Ics, Line, Tlen+1, Action, Alen);
yystate(108, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,108};
yystate(107, [116|Ics], Line, Tlen, _, _) ->
    yystate(103, Ics, Line, Tlen+1, 50, Tlen);
yystate(107, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(107, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(107, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(107, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(107, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(107, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(107, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(107, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,107};
yystate(106, [101|Ics], Line, Tlen, _, _) ->
    yystate(102, Ics, Line, Tlen+1, 50, Tlen);
yystate(106, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(106, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(106, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(106, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(106, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(106, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(106, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(106, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,106};
yystate(105, [114|Ics], Line, Tlen, _, _) ->
    yystate(109, Ics, Line, Tlen+1, 50, Tlen);
yystate(105, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(105, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(105, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(105, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(105, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(105, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(105, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(105, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,105};
yystate(104, Ics, Line, Tlen, _, _) ->
    {54,Tlen,Ics,Line};
yystate(103, [97|Ics], Line, Tlen, _, _) ->
    yystate(99, Ics, Line, Tlen+1, 50, Tlen);
yystate(103, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(103, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(103, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(103, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(103, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(103, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(103, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,103};
yystate(102, [110|Ics], Line, Tlen, _, _) ->
    yystate(98, Ics, Line, Tlen+1, 50, Tlen);
yystate(102, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(102, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(102, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(102, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(102, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(102, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(102, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(102, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,102};
yystate(101, [116|Ics], Line, Tlen, _, _) ->
    yystate(105, Ics, Line, Tlen+1, 50, Tlen);
yystate(101, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(101, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(101, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(101, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(101, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(101, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(101, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(101, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,101};
yystate(100, [125|Ics], Line, Tlen, Action, Alen) ->
    yystate(100, Ics, Line, Tlen+1, Action, Alen);
yystate(100, [42|Ics], Line, Tlen, Action, Alen) ->
    yystate(104, Ics, Line, Tlen+1, Action, Alen);
yystate(100, [10|Ics], Line, Tlen, Action, Alen) ->
    yystate(108, Ics, Line+1, Tlen+1, Action, Alen);
yystate(100, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(108, Ics, Line, Tlen+1, Action, Alen);
yystate(100, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 41 ->
    yystate(108, Ics, Line, Tlen+1, Action, Alen);
yystate(100, [C|Ics], Line, Tlen, Action, Alen) when C >= 43, C =< 124 ->
    yystate(108, Ics, Line, Tlen+1, Action, Alen);
yystate(100, [C|Ics], Line, Tlen, Action, Alen) when C >= 126 ->
    yystate(108, Ics, Line, Tlen+1, Action, Alen);
yystate(100, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,100};
yystate(99, [118|Ics], Line, Tlen, _, _) ->
    yystate(95, Ics, Line, Tlen+1, 50, Tlen);
yystate(99, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(99, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(99, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(99, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(99, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(99, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 117 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(99, [C|Ics], Line, Tlen, _, _) when C >= 119, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(99, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,99};
yystate(98, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 46, Tlen);
yystate(98, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 46, Tlen);
yystate(98, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 46, Tlen);
yystate(98, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 46, Tlen);
yystate(98, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 46, Tlen);
yystate(98, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 46, Tlen);
yystate(98, Ics, Line, Tlen, _, _) ->
    {46,Tlen,Ics,Line,98};
yystate(97, Ics, Line, Tlen, _, _) ->
    {36,Tlen,Ics,Line};
yystate(96, [123|Ics], Line, Tlen, Action, Alen) ->
    yystate(108, Ics, Line, Tlen+1, Action, Alen);
yystate(96, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,96};
yystate(95, [101|Ics], Line, Tlen, _, _) ->
    yystate(91, Ics, Line, Tlen+1, 50, Tlen);
yystate(95, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(95, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(95, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(95, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(95, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(95, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(95, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(95, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,95};
yystate(94, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(94, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(94, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(94, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(94, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(94, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(94, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,94};
yystate(93, Ics, Line, Tlen, _, _) ->
    {43,Tlen,Ics,Line};
yystate(92, Ics, Line, Tlen, _, _) ->
    {39,Tlen,Ics,Line};
yystate(91, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 4, Tlen);
yystate(91, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 4, Tlen);
yystate(91, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 4, Tlen);
yystate(91, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 4, Tlen);
yystate(91, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 4, Tlen);
yystate(91, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 4, Tlen);
yystate(91, Ics, Line, Tlen, _, _) ->
    {4,Tlen,Ics,Line,91};
yystate(90, [114|Ics], Line, Tlen, _, _) ->
    yystate(86, Ics, Line, Tlen+1, 50, Tlen);
yystate(90, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(90, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(90, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(90, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(90, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(90, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(90, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(90, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,90};
yystate(89, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 49, Tlen);
yystate(89, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 49, Tlen);
yystate(89, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 49, Tlen);
yystate(89, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 49, Tlen);
yystate(89, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 49, Tlen);
yystate(89, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 49, Tlen);
yystate(89, Ics, Line, Tlen, _, _) ->
    {49,Tlen,Ics,Line,89};
yystate(88, Ics, Line, Tlen, _, _) ->
    {19,Tlen,Ics,Line};
yystate(87, [116|Ics], Line, Tlen, _, _) ->
    yystate(83, Ics, Line, Tlen+1, 50, Tlen);
yystate(87, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(87, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(87, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(87, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(87, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(87, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(87, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(87, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,87};
yystate(86, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 48, Tlen);
yystate(86, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 48, Tlen);
yystate(86, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 48, Tlen);
yystate(86, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 48, Tlen);
yystate(86, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 48, Tlen);
yystate(86, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 48, Tlen);
yystate(86, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,86};
yystate(85, [100|Ics], Line, Tlen, _, _) ->
    yystate(89, Ics, Line, Tlen+1, 50, Tlen);
yystate(85, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(85, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(85, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(85, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(85, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(85, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(85, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(85, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,85};
yystate(84, Ics, Line, Tlen, _, _) ->
    {40,Tlen,Ics,Line};
yystate(83, [118|Ics], Line, Tlen, _, _) ->
    yystate(79, Ics, Line, Tlen+1, 50, Tlen);
yystate(83, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(83, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(83, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(83, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(83, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(83, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 117 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(83, [C|Ics], Line, Tlen, _, _) when C >= 119, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(83, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,83};
yystate(82, [111|Ics], Line, Tlen, _, _) ->
    yystate(78, Ics, Line, Tlen+1, 50, Tlen);
yystate(82, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(82, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(82, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(82, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(82, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(82, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(82, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(82, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,82};
yystate(81, [110|Ics], Line, Tlen, _, _) ->
    yystate(85, Ics, Line, Tlen+1, 50, Tlen);
yystate(81, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(81, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(81, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(81, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(81, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(81, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(81, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(81, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,81};
yystate(80, [62|Ics], Line, Tlen, Action, Alen) ->
    yystate(84, Ics, Line, Tlen+1, Action, Alen);
yystate(80, [48|Ics], Line, Tlen, Action, Alen) ->
    yystate(76, Ics, Line, Tlen+1, Action, Alen);
yystate(80, [C|Ics], Line, Tlen, Action, Alen) when C >= 49, C =< 57 ->
    yystate(72, Ics, Line, Tlen+1, Action, Alen);
yystate(80, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,80};
yystate(79, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 12, Tlen);
yystate(79, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 12, Tlen);
yystate(79, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 12, Tlen);
yystate(79, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 12, Tlen);
yystate(79, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 12, Tlen);
yystate(79, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 12, Tlen);
yystate(79, Ics, Line, Tlen, _, _) ->
    {12,Tlen,Ics,Line,79};
yystate(78, [116|Ics], Line, Tlen, _, _) ->
    yystate(74, Ics, Line, Tlen+1, 50, Tlen);
yystate(78, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(78, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(78, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(78, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(78, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(78, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(78, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(78, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,78};
yystate(77, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 20, Tlen);
yystate(77, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 20, Tlen);
yystate(77, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 20, Tlen);
yystate(77, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 20, Tlen);
yystate(77, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 20, Tlen);
yystate(77, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 20, Tlen);
yystate(77, Ics, Line, Tlen, _, _) ->
    {20,Tlen,Ics,Line,77};
yystate(76, Ics, Line, Tlen, _, _) ->
    {51,Tlen,Ics,Line};
yystate(75, [97|Ics], Line, Tlen, _, _) ->
    yystate(71, Ics, Line, Tlen+1, 50, Tlen);
yystate(75, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(75, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(75, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(75, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(75, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(75, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(75, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,75};
yystate(74, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 38, Tlen);
yystate(74, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 38, Tlen);
yystate(74, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 38, Tlen);
yystate(74, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 38, Tlen);
yystate(74, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 38, Tlen);
yystate(74, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 38, Tlen);
yystate(74, Ics, Line, Tlen, _, _) ->
    {38,Tlen,Ics,Line,74};
yystate(73, [102|Ics], Line, Tlen, _, _) ->
    yystate(77, Ics, Line, Tlen+1, 50, Tlen);
yystate(73, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(73, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(73, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(73, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(73, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(73, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 101 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(73, [C|Ics], Line, Tlen, _, _) when C >= 103, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(73, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,73};
yystate(72, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(72, Ics, Line, Tlen+1, 51, Tlen);
yystate(72, Ics, Line, Tlen, _, _) ->
    {51,Tlen,Ics,Line,72};
yystate(71, [116|Ics], Line, Tlen, _, _) ->
    yystate(67, Ics, Line, Tlen+1, 50, Tlen);
yystate(71, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(71, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(71, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(71, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(71, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(71, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(71, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(71, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,71};
yystate(70, [101|Ics], Line, Tlen, _, _) ->
    yystate(66, Ics, Line, Tlen+1, 50, Tlen);
yystate(70, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(70, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(70, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(70, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(70, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(70, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(70, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(70, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,70};
yystate(69, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 21, Tlen);
yystate(69, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 21, Tlen);
yystate(69, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 21, Tlen);
yystate(69, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 21, Tlen);
yystate(69, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 21, Tlen);
yystate(69, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 21, Tlen);
yystate(69, Ics, Line, Tlen, _, _) ->
    {21,Tlen,Ics,Line,69};
yystate(68, Ics, Line, Tlen, _, _) ->
    {18,Tlen,Ics,Line};
yystate(67, [108|Ics], Line, Tlen, _, _) ->
    yystate(63, Ics, Line, Tlen+1, 50, Tlen);
yystate(67, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(67, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(67, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(67, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(67, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(67, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(67, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(67, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,67};
yystate(66, [116|Ics], Line, Tlen, _, _) ->
    yystate(62, Ics, Line, Tlen+1, 50, Tlen);
yystate(66, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(66, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(66, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(66, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(66, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(66, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(66, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(66, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,66};
yystate(65, [111|Ics], Line, Tlen, _, _) ->
    yystate(69, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [101|Ics], Line, Tlen, _, _) ->
    yystate(73, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 110 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(65, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,65};
yystate(64, Ics, Line, Tlen, _, _) ->
    {45,Tlen,Ics,Line};
yystate(63, [97|Ics], Line, Tlen, _, _) ->
    yystate(59, Ics, Line, Tlen+1, 50, Tlen);
yystate(63, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(63, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(63, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(63, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(63, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(63, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(63, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,63};
yystate(62, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 14, Tlen);
yystate(62, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 14, Tlen);
yystate(62, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 14, Tlen);
yystate(62, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 14, Tlen);
yystate(62, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 14, Tlen);
yystate(62, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 14, Tlen);
yystate(62, Ics, Line, Tlen, _, _) ->
    {14,Tlen,Ics,Line,62};
yystate(61, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 23, Tlen);
yystate(61, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 23, Tlen);
yystate(61, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 23, Tlen);
yystate(61, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 23, Tlen);
yystate(61, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 23, Tlen);
yystate(61, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 23, Tlen);
yystate(61, Ics, Line, Tlen, _, _) ->
    {23,Tlen,Ics,Line,61};
yystate(60, Ics, Line, Tlen, _, _) ->
    {33,Tlen,Ics,Line};
yystate(59, [98|Ics], Line, Tlen, _, _) ->
    yystate(55, Ics, Line, Tlen+1, 50, Tlen);
yystate(59, [97|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(59, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(59, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(59, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(59, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(59, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(59, [C|Ics], Line, Tlen, _, _) when C >= 99, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(59, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,59};
yystate(58, [115|Ics], Line, Tlen, _, _) ->
    yystate(54, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [110|Ics], Line, Tlen, _, _) ->
    yystate(38, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [109|Ics], Line, Tlen, _, _) ->
    yystate(34, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [102|Ics], Line, Tlen, _, _) ->
    yystate(14, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 101 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [C|Ics], Line, Tlen, _, _) when C >= 103, C =< 108 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 114 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(58, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,58};
yystate(57, [101|Ics], Line, Tlen, _, _) ->
    yystate(61, Ics, Line, Tlen+1, 50, Tlen);
yystate(57, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(57, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(57, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(57, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(57, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(57, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(57, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(57, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,57};
yystate(56, [45|Ics], Line, Tlen, _, _) ->
    yystate(60, Ics, Line, Tlen+1, 37, Tlen);
yystate(56, Ics, Line, Tlen, _, _) ->
    {37,Tlen,Ics,Line,56};
yystate(55, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 3, Tlen);
yystate(55, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 3, Tlen);
yystate(55, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 3, Tlen);
yystate(55, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 3, Tlen);
yystate(55, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 3, Tlen);
yystate(55, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 3, Tlen);
yystate(55, Ics, Line, Tlen, _, _) ->
    {3,Tlen,Ics,Line,55};
yystate(54, [110|Ics], Line, Tlen, _, _) ->
    yystate(50, Ics, Line, Tlen+1, 50, Tlen);
yystate(54, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(54, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(54, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(54, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(54, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(54, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(54, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(54, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,54};
yystate(53, [115|Ics], Line, Tlen, _, _) ->
    yystate(57, Ics, Line, Tlen+1, 50, Tlen);
yystate(53, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(53, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(53, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(53, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,53};
yystate(52, Ics, Line, Tlen, _, _) ->
    {16,Tlen,Ics,Line};
yystate(51, [97|Ics], Line, Tlen, _, _) ->
    yystate(47, Ics, Line, Tlen+1, 50, Tlen);
yystate(51, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(51, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(51, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(51, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(51, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(51, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(51, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,51};
yystate(50, [105|Ics], Line, Tlen, _, _) ->
    yystate(46, Ics, Line, Tlen+1, 50, Tlen);
yystate(50, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(50, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(50, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(50, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(50, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(50, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(50, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(50, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,50};
yystate(49, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 25, Tlen);
yystate(49, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 25, Tlen);
yystate(49, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 25, Tlen);
yystate(49, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 25, Tlen);
yystate(49, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 25, Tlen);
yystate(49, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 25, Tlen);
yystate(49, Ics, Line, Tlen, _, _) ->
    {25,Tlen,Ics,Line,49};
yystate(48, [61|Ics], Line, Tlen, _, _) ->
    yystate(52, Ics, Line, Tlen+1, 26, Tlen);
yystate(48, Ics, Line, Tlen, _, _) ->
    {26,Tlen,Ics,Line,48};
yystate(47, [118|Ics], Line, Tlen, _, _) ->
    yystate(43, Ics, Line, Tlen+1, 50, Tlen);
yystate(47, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(47, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(47, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(47, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(47, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(47, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 117 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(47, [C|Ics], Line, Tlen, _, _) when C >= 119, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(47, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,47};
yystate(46, [108|Ics], Line, Tlen, _, _) ->
    yystate(42, Ics, Line, Tlen+1, 50, Tlen);
yystate(46, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(46, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(46, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(46, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(46, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(46, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(46, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(46, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,46};
yystate(45, [100|Ics], Line, Tlen, _, _) ->
    yystate(49, Ics, Line, Tlen+1, 50, Tlen);
yystate(45, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(45, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(45, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(45, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(45, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(45, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(45, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(45, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,45};
yystate(44, Ics, Line, Tlen, _, _) ->
    {22,Tlen,Ics,Line};
yystate(43, [97|Ics], Line, Tlen, _, _) ->
    yystate(39, Ics, Line, Tlen+1, 50, Tlen);
yystate(43, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(43, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(43, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(43, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(43, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(43, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(43, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,43};
yystate(42, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 32, Tlen);
yystate(42, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 32, Tlen);
yystate(42, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 32, Tlen);
yystate(42, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 32, Tlen);
yystate(42, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 32, Tlen);
yystate(42, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 32, Tlen);
yystate(42, Ics, Line, Tlen, _, _) ->
    {32,Tlen,Ics,Line,42};
yystate(41, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 24, Tlen);
yystate(41, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 24, Tlen);
yystate(41, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 24, Tlen);
yystate(41, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 24, Tlen);
yystate(41, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 24, Tlen);
yystate(41, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 24, Tlen);
yystate(41, Ics, Line, Tlen, _, _) ->
    {24,Tlen,Ics,Line,41};
yystate(40, [62|Ics], Line, Tlen, _, _) ->
    yystate(44, Ics, Line, Tlen+1, 44, Tlen);
yystate(40, Ics, Line, Tlen, _, _) ->
    {44,Tlen,Ics,Line,40};
yystate(39, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 2, Tlen);
yystate(39, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 2, Tlen);
yystate(39, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 2, Tlen);
yystate(39, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 2, Tlen);
yystate(39, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 2, Tlen);
yystate(39, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 2, Tlen);
yystate(39, Ics, Line, Tlen, _, _) ->
    {2,Tlen,Ics,Line,39};
yystate(38, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 31, Tlen);
yystate(38, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 31, Tlen);
yystate(38, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 31, Tlen);
yystate(38, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 31, Tlen);
yystate(38, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 31, Tlen);
yystate(38, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 31, Tlen);
yystate(38, Ics, Line, Tlen, _, _) ->
    {31,Tlen,Ics,Line,38};
yystate(37, [114|Ics], Line, Tlen, _, _) ->
    yystate(41, Ics, Line, Tlen+1, 50, Tlen);
yystate(37, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(37, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(37, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(37, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(37, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(37, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(37, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(37, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,37};
yystate(36, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 0, Tlen);
yystate(36, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 0, Tlen);
yystate(36, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 0, Tlen);
yystate(36, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 0, Tlen);
yystate(36, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 0, Tlen);
yystate(36, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 0, Tlen);
yystate(36, Ics, Line, Tlen, _, _) ->
    {0,Tlen,Ics,Line,36};
yystate(35, [114|Ics], Line, Tlen, _, _) ->
    yystate(31, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [105|Ics], Line, Tlen, _, _) ->
    yystate(23, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(35, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,35};
yystate(34, [112|Ics], Line, Tlen, _, _) ->
    yystate(30, Ics, Line, Tlen+1, 50, Tlen);
yystate(34, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(34, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(34, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(34, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(34, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(34, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 111 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(34, [C|Ics], Line, Tlen, _, _) when C >= 113, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(34, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,34};
yystate(33, [111|Ics], Line, Tlen, _, _) ->
    yystate(37, Ics, Line, Tlen+1, 50, Tlen);
yystate(33, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(33, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(33, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(33, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(33, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(33, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(33, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(33, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,33};
yystate(32, [104|Ics], Line, Tlen, _, _) ->
    yystate(36, Ics, Line, Tlen+1, 50, Tlen);
yystate(32, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(32, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(32, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(32, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(32, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(32, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 103 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(32, [C|Ics], Line, Tlen, _, _) when C >= 105, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(32, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,32};
yystate(31, [110|Ics], Line, Tlen, _, _) ->
    yystate(27, Ics, Line, Tlen+1, 50, Tlen);
yystate(31, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(31, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(31, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(31, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(31, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(31, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(31, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(31, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,31};
yystate(30, [111|Ics], Line, Tlen, _, _) ->
    yystate(26, Ics, Line, Tlen+1, 50, Tlen);
yystate(30, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(30, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(30, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(30, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(30, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(30, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(30, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(30, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,30};
yystate(29, [114|Ics], Line, Tlen, _, _) ->
    yystate(33, Ics, Line, Tlen+1, 50, Tlen);
yystate(29, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(29, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(29, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(29, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(29, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(29, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(29, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(29, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,29};
yystate(28, [115|Ics], Line, Tlen, _, _) ->
    yystate(32, Ics, Line, Tlen+1, 50, Tlen);
yystate(28, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(28, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(28, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(28, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(28, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(28, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(28, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(28, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,28};
yystate(27, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 13, Tlen);
yystate(27, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 13, Tlen);
yystate(27, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 13, Tlen);
yystate(27, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 13, Tlen);
yystate(27, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 13, Tlen);
yystate(27, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 13, Tlen);
yystate(27, Ics, Line, Tlen, _, _) ->
    {13,Tlen,Ics,Line,27};
yystate(26, [114|Ics], Line, Tlen, _, _) ->
    yystate(22, Ics, Line, Tlen+1, 50, Tlen);
yystate(26, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(26, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(26, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(26, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(26, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(26, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(26, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(26, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,26};
yystate(25, [114|Ics], Line, Tlen, _, _) ->
    yystate(29, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [110|Ics], Line, Tlen, _, _) ->
    yystate(45, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [109|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [108|Ics], Line, Tlen, _, _) ->
    yystate(53, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(25, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,25};
yystate(24, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 11, Tlen);
yystate(24, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 11, Tlen);
yystate(24, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 11, Tlen);
yystate(24, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 11, Tlen);
yystate(24, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 11, Tlen);
yystate(24, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 11, Tlen);
yystate(24, Ics, Line, Tlen, _, _) ->
    {11,Tlen,Ics,Line,24};
yystate(23, [108|Ics], Line, Tlen, _, _) ->
    yystate(19, Ics, Line, Tlen+1, 50, Tlen);
yystate(23, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(23, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(23, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(23, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(23, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(23, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(23, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(23, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,23};
yystate(22, [116|Ics], Line, Tlen, _, _) ->
    yystate(18, Ics, Line, Tlen+1, 50, Tlen);
yystate(22, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(22, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(22, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(22, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(22, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(22, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(22, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(22, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,22};
yystate(21, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, Ics, Line, Tlen, _, _) ->
    {27,Tlen,Ics,Line,21};
yystate(20, [108|Ics], Line, Tlen, _, _) ->
    yystate(24, Ics, Line, Tlen+1, 50, Tlen);
yystate(20, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(20, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(20, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(20, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(20, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(20, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(20, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(20, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,20};
yystate(19, [101|Ics], Line, Tlen, _, _) ->
    yystate(15, Ics, Line, Tlen+1, 50, Tlen);
yystate(19, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(19, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(19, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(19, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(19, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(19, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(19, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(19, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,19};
yystate(18, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 30, Tlen);
yystate(18, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 30, Tlen);
yystate(18, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 30, Tlen);
yystate(18, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 30, Tlen);
yystate(18, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 30, Tlen);
yystate(18, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 30, Tlen);
yystate(18, Ics, Line, Tlen, _, _) ->
    {30,Tlen,Ics,Line,18};
yystate(17, [101|Ics], Line, Tlen, _, _) ->
    yystate(21, Ics, Line, Tlen+1, 50, Tlen);
yystate(17, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(17, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(17, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(17, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(17, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(17, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(17, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(17, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,17};
yystate(16, [111|Ics], Line, Tlen, _, _) ->
    yystate(20, Ics, Line, Tlen+1, 50, Tlen);
yystate(16, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(16, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(16, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(16, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(16, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(16, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(16, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(16, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,16};
yystate(15, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 10, Tlen);
yystate(15, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 10, Tlen);
yystate(15, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 10, Tlen);
yystate(15, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 10, Tlen);
yystate(15, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 10, Tlen);
yystate(15, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 10, Tlen);
yystate(15, Ics, Line, Tlen, _, _) ->
    {10,Tlen,Ics,Line,15};
yystate(14, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 17, Tlen);
yystate(14, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 17, Tlen);
yystate(14, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 17, Tlen);
yystate(14, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 17, Tlen);
yystate(14, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 17, Tlen);
yystate(14, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 17, Tlen);
yystate(14, Ics, Line, Tlen, _, _) ->
    {17,Tlen,Ics,Line,14};
yystate(13, [115|Ics], Line, Tlen, _, _) ->
    yystate(17, Ics, Line, Tlen+1, 50, Tlen);
yystate(13, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(13, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(13, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(13, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(13, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(13, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(13, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(13, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,13};
yystate(12, [111|Ics], Line, Tlen, _, _) ->
    yystate(16, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, [97|Ics], Line, Tlen, _, _) ->
    yystate(28, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 110 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(12, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,12};
yystate(11, [114|Ics], Line, Tlen, _, _) ->
    yystate(7, Ics, Line, Tlen+1, 50, Tlen);
yystate(11, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(11, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(11, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(11, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(11, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(11, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(11, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(11, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,11};
yystate(10, [111|Ics], Line, Tlen, _, _) ->
    yystate(6, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, [97|Ics], Line, Tlen, _, _) ->
    yystate(9, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 110 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(10, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,10};
yystate(9, [108|Ics], Line, Tlen, _, _) ->
    yystate(13, Ics, Line, Tlen+1, 50, Tlen);
yystate(9, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(9, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(9, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(9, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(9, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(9, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(9, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(9, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,9};
yystate(8, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 1, Tlen);
yystate(8, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 1, Tlen);
yystate(8, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 1, Tlen);
yystate(8, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 1, Tlen);
yystate(8, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 1, Tlen);
yystate(8, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 1, Tlen);
yystate(8, Ics, Line, Tlen, _, _) ->
    {1,Tlen,Ics,Line,8};
yystate(7, [108|Ics], Line, Tlen, _, _) ->
    yystate(3, Ics, Line, Tlen+1, 50, Tlen);
yystate(7, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(7, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(7, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(7, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(7, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(7, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(7, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(7, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,7};
yystate(6, [114|Ics], Line, Tlen, _, _) ->
    yystate(2, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [108|Ics], Line, Tlen, _, _) ->
    yystate(1, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 113 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(6, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,6};
yystate(5, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 28, Tlen);
yystate(5, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 28, Tlen);
yystate(5, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 28, Tlen);
yystate(5, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 28, Tlen);
yystate(5, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 28, Tlen);
yystate(5, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 28, Tlen);
yystate(5, Ics, Line, Tlen, _, _) ->
    {28,Tlen,Ics,Line,5};
yystate(4, [103|Ics], Line, Tlen, _, _) ->
    yystate(8, Ics, Line, Tlen+1, 50, Tlen);
yystate(4, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(4, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(4, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(4, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(4, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(4, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 102 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(4, [C|Ics], Line, Tlen, _, _) when C >= 104, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(4, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,4};
yystate(3, [97|Ics], Line, Tlen, _, _) ->
    yystate(0, Ics, Line, Tlen+1, 50, Tlen);
yystate(3, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(3, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(3, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(3, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(3, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(3, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(3, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,3};
yystate(2, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 29, Tlen);
yystate(2, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 29, Tlen);
yystate(2, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 29, Tlen);
yystate(2, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 29, Tlen);
yystate(2, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 29, Tlen);
yystate(2, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 29, Tlen);
yystate(2, Ics, Line, Tlen, _, _) ->
    {29,Tlen,Ics,Line,2};
yystate(1, [100|Ics], Line, Tlen, _, _) ->
    yystate(5, Ics, Line, Tlen+1, 50, Tlen);
yystate(1, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(1, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(1, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(1, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(1, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(1, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(1, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(1, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,1};
yystate(0, [110|Ics], Line, Tlen, _, _) ->
    yystate(4, Ics, Line, Tlen+1, 50, Tlen);
yystate(0, [95|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(0, [45|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(0, [46|Ics], Line, Tlen, _, _) ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(0, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(0, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(0, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(0, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(94, Ics, Line, Tlen+1, 50, Tlen);
yystate(0, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line,0};
yystate(S, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,S}.

%% yyaction(Action, TokenLength, TokenChars, TokenLine) ->
%% {token,Token} | {end_token, Token} | skip_token | {error,String}.
%% Generated action function.

yyaction(0, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_0(TokenChars, TokenLine);
yyaction(1, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_1(TokenChars, TokenLine);
yyaction(2, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_2(TokenChars, TokenLine);
yyaction(3, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_3(TokenChars, TokenLine);
yyaction(4, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_4(TokenChars, TokenLine);
yyaction(5, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_5(TokenChars, TokenLine);
yyaction(6, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_6(TokenChars, TokenLine);
yyaction(7, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_7(TokenChars, TokenLine);
yyaction(8, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_8(TokenChars, TokenLine);
yyaction(9, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_9(TokenChars, TokenLine);
yyaction(10, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_10(TokenChars, TokenLine);
yyaction(11, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_11(TokenChars, TokenLine);
yyaction(12, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_12(TokenChars, TokenLine);
yyaction(13, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_13(TokenChars, TokenLine);
yyaction(14, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_14(TokenChars, TokenLine);
yyaction(15, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_15(TokenChars, TokenLine);
yyaction(16, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_16(TokenChars, TokenLine);
yyaction(17, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_17(TokenChars, TokenLine);
yyaction(18, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_18(TokenChars, TokenLine);
yyaction(19, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_19(TokenChars, TokenLine);
yyaction(20, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_20(TokenChars, TokenLine);
yyaction(21, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_21(TokenChars, TokenLine);
yyaction(22, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_22(TokenChars, TokenLine);
yyaction(23, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_23(TokenChars, TokenLine);
yyaction(24, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_24(TokenChars, TokenLine);
yyaction(25, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_25(TokenChars, TokenLine);
yyaction(26, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_26(TokenChars, TokenLine);
yyaction(27, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_27(TokenChars, TokenLine);
yyaction(28, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_28(TokenChars, TokenLine);
yyaction(29, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_29(TokenChars, TokenLine);
yyaction(30, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_30(TokenChars, TokenLine);
yyaction(31, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_31(TokenChars, TokenLine);
yyaction(32, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_32(TokenChars, TokenLine);
yyaction(33, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_33(TokenChars, TokenLine);
yyaction(34, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_34(TokenChars, TokenLine);
yyaction(35, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_35(TokenChars, TokenLine);
yyaction(36, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_36(TokenChars, TokenLine);
yyaction(37, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_37(TokenChars, TokenLine);
yyaction(38, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_38(TokenChars, TokenLine);
yyaction(39, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_39(TokenChars, TokenLine);
yyaction(40, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_40(TokenChars, TokenLine);
yyaction(41, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_41(TokenChars, TokenLine);
yyaction(42, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_42(TokenChars, TokenLine);
yyaction(43, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_43(TokenChars, TokenLine);
yyaction(44, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_44(TokenChars, TokenLine);
yyaction(45, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_45(TokenChars, TokenLine);
yyaction(46, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_46(TokenChars, TokenLine);
yyaction(47, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_47(TokenChars, TokenLine);
yyaction(48, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_48(TokenChars, TokenLine);
yyaction(49, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_49(TokenChars, TokenLine);
yyaction(50, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_50(TokenChars, TokenLine);
yyaction(51, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_51(TokenChars, TokenLine);
yyaction(52, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_52(TokenChars, TokenLine);
yyaction(53, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_53(TokenChars, TokenLine);
yyaction(54, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_54(TokenChars, TokenLine);
yyaction(55, _, _, _) ->
    yyaction_55();
yyaction(56, _, _, _) ->
    yyaction_56();
yyaction(_, _, _, _) -> error.

-compile({inline,yyaction_0/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 105).
yyaction_0(TokenChars, TokenLine) ->
     { token, { l_bash, TokenLine, TokenChars } } .

-compile({inline,yyaction_1/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 106).
yyaction_1(TokenChars, TokenLine) ->
     { token, { l_erlang, TokenLine, TokenChars } } .

-compile({inline,yyaction_2/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 107).
yyaction_2(TokenChars, TokenLine) ->
     { token, { l_java, TokenLine, TokenChars } } .

-compile({inline,yyaction_3/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 108).
yyaction_3(TokenChars, TokenLine) ->
     { token, { l_matlab, TokenLine, TokenChars } } .

-compile({inline,yyaction_4/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 109).
yyaction_4(TokenChars, TokenLine) ->
     { token, { l_octave, TokenLine, TokenChars } } .

-compile({inline,yyaction_5/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 110).
yyaction_5(TokenChars, TokenLine) ->
     { token, { l_perl, TokenLine, TokenChars } } .

-compile({inline,yyaction_6/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 111).
yyaction_6(TokenChars, TokenLine) ->
     { token, { l_python, TokenLine, TokenChars } } .

-compile({inline,yyaction_7/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 112).
yyaction_7(TokenChars, TokenLine) ->
     { token, { l_r, TokenLine, TokenChars } } .

-compile({inline,yyaction_8/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 113).
yyaction_8(TokenChars, TokenLine) ->
     { token, { l_racket, TokenLine, TokenChars } } .

-compile({inline,yyaction_9/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 115).
yyaction_9(TokenChars, TokenLine) ->
     { token, { t_str, TokenLine, TokenChars } } .

-compile({inline,yyaction_10/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 116).
yyaction_10(TokenChars, TokenLine) ->
     { token, { t_file, TokenLine, TokenChars } } .

-compile({inline,yyaction_11/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 117).
yyaction_11(TokenChars, TokenLine) ->
     { token, { t_bool, TokenLine, TokenChars } } .

-compile({inline,yyaction_12/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 118).
yyaction_12(TokenChars, TokenLine) ->
     { token, { t_fn_ntv, TokenLine, TokenChars } } .

-compile({inline,yyaction_13/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 119).
yyaction_13(TokenChars, TokenLine) ->
     { token, { t_fn_frn, TokenLine, TokenChars } } .

-compile({inline,yyaction_14/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 121).
yyaction_14(TokenChars, TokenLine) ->
     { token, { assign, TokenLine, TokenChars } } .

-compile({inline,yyaction_15/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 122).
yyaction_15(TokenChars, TokenLine) ->
     { token, { bar, TokenLine, TokenChars } } .

-compile({inline,yyaction_16/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 123).
yyaction_16(TokenChars, TokenLine) ->
     { token, { cmp, TokenLine, TokenChars } } .

-compile({inline,yyaction_17/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 124).
yyaction_17(TokenChars, TokenLine) ->
     { token, { cnd, TokenLine, TokenChars } } .

-compile({inline,yyaction_18/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 125).
yyaction_18(TokenChars, TokenLine) ->
     { token, { colon, TokenLine, TokenChars } } .

-compile({inline,yyaction_19/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 126).
yyaction_19(TokenChars, TokenLine) ->
     { token, { comma, TokenLine, TokenChars } } .

-compile({inline,yyaction_20/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 127).
yyaction_20(TokenChars, TokenLine) ->
     { token, { def, TokenLine, TokenChars } } .

-compile({inline,yyaction_21/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 128).
yyaction_21(TokenChars, TokenLine) ->
     { token, { do, TokenLine, TokenChars } } .

-compile({inline,yyaction_22/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 129).
yyaction_22(TokenChars, TokenLine) ->
     { token, { doublertag, TokenLine, TokenChars } } .

-compile({inline,yyaction_23/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 130).
yyaction_23(TokenChars, TokenLine) ->
     { token, { else, TokenLine, TokenChars } } .

-compile({inline,yyaction_24/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 131).
yyaction_24(TokenChars, TokenLine) ->
     { token, { err, TokenLine, TokenChars } } .

-compile({inline,yyaction_25/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 132).
yyaction_25(TokenChars, TokenLine) ->
     { token, { halt, TokenLine, TokenChars } } .

-compile({inline,yyaction_26/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 133).
yyaction_26(TokenChars, TokenLine) ->
     { token, { eq, TokenLine, TokenChars } } .

-compile({inline,yyaction_27/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 134).
yyaction_27(TokenChars, TokenLine) ->
     { token, { false, TokenLine, TokenChars } } .

-compile({inline,yyaction_28/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 135).
yyaction_28(TokenChars, TokenLine) ->
     { token, { fold, TokenLine, TokenChars } } .

-compile({inline,yyaction_29/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 136).
yyaction_29(TokenChars, TokenLine) ->
     { token, { for, TokenLine, TokenChars } } .

-compile({inline,yyaction_30/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 137).
yyaction_30(TokenChars, TokenLine) ->
     { token, { import, TokenLine, TokenChars } } .

-compile({inline,yyaction_31/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 138).
yyaction_31(TokenChars, TokenLine) ->
     { token, { in, TokenLine, TokenChars } } .

-compile({inline,yyaction_32/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 139).
yyaction_32(TokenChars, TokenLine) ->
     { token, { isnil, TokenLine, TokenChars } } .

-compile({inline,yyaction_33/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 140).
yyaction_33(TokenChars, TokenLine) ->
     { token, { larrow, TokenLine, TokenChars } } .

-compile({inline,yyaction_34/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 141).
yyaction_34(TokenChars, TokenLine) ->
     { token, { lbrace, TokenLine, TokenChars } } .

-compile({inline,yyaction_35/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 142).
yyaction_35(TokenChars, TokenLine) ->
     { token, { lparen, TokenLine, TokenChars } } .

-compile({inline,yyaction_36/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 143).
yyaction_36(TokenChars, TokenLine) ->
     { token, { lsquarebr, TokenLine, TokenChars } } .

-compile({inline,yyaction_37/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 144).
yyaction_37(TokenChars, TokenLine) ->
     { token, { ltag, TokenLine, TokenChars } } .

-compile({inline,yyaction_38/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 145).
yyaction_38(TokenChars, TokenLine) ->
     { token, { neg, TokenLine, TokenChars } } .

-compile({inline,yyaction_39/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 146).
yyaction_39(TokenChars, TokenLine) ->
     { token, { plus, TokenLine, TokenChars } } .

-compile({inline,yyaction_40/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 147).
yyaction_40(TokenChars, TokenLine) ->
     { token, { rarrow, TokenLine, TokenChars } } .

-compile({inline,yyaction_41/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 148).
yyaction_41(TokenChars, TokenLine) ->
     { token, { rbrace, TokenLine, TokenChars } } .

-compile({inline,yyaction_42/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 149).
yyaction_42(TokenChars, TokenLine) ->
     { token, { rparen, TokenLine, TokenChars } } .

-compile({inline,yyaction_43/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 150).
yyaction_43(TokenChars, TokenLine) ->
     { token, { rsquarebr, TokenLine, TokenChars } } .

-compile({inline,yyaction_44/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 151).
yyaction_44(TokenChars, TokenLine) ->
     { token, { rtag, TokenLine, TokenChars } } .

-compile({inline,yyaction_45/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 152).
yyaction_45(TokenChars, TokenLine) ->
     { token, { semicolon, TokenLine, TokenChars } } .

-compile({inline,yyaction_46/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 153).
yyaction_46(TokenChars, TokenLine) ->
     { token, { then, TokenLine, TokenChars } } .

-compile({inline,yyaction_47/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 154).
yyaction_47(TokenChars, TokenLine) ->
     { token, { true, TokenLine, TokenChars } } .

-compile({inline,yyaction_48/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 155).
yyaction_48(TokenChars, TokenLine) ->
     { token, { vee, TokenLine, TokenChars } } .

-compile({inline,yyaction_49/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 156).
yyaction_49(TokenChars, TokenLine) ->
     { token, { wedge, TokenLine, TokenChars } } .

-compile({inline,yyaction_50/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 158).
yyaction_50(TokenChars, TokenLine) ->
     { token, { id, TokenLine, TokenChars } } .

-compile({inline,yyaction_51/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 159).
yyaction_51(TokenChars, TokenLine) ->
     { token, { intlit, TokenLine, TokenChars } } .

-compile({inline,yyaction_52/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 160).
yyaction_52(TokenChars, TokenLine) ->
     { token, { strlit, TokenLine, trim_lit (TokenChars) } } .

-compile({inline,yyaction_53/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 161).
yyaction_53(TokenChars, TokenLine) ->
     { token, { filelit, TokenLine, trim_lit (TokenChars) } } .

-compile({inline,yyaction_54/2}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 162).
yyaction_54(TokenChars, TokenLine) ->
     { token, { body, TokenLine, trim_body (TokenChars) } } .

-compile({inline,yyaction_55/0}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 164).
yyaction_55() ->
     skip_token .

-compile({inline,yyaction_56/0}).
-file("/home/jorgen/git/cf_client/_build/test/lib/cf_client/src/cuneiform_scan.xrl", 165).
yyaction_56() ->
     skip_token .

-file("/usr/local/lib/erlang/lib/parsetools-2.1.6/include/leexinc.hrl", 313).
